#include <sys/types.h>
#include <stdio.h>
#include <string.h>

#include <CL/cl.h>

#define KERNEL_ITERATIONS 256
#define MAX_SMI 16
#define MAX_BLOCK_SIZE 1024

#ifndef SHMEM
# define SHMEM 1
#endif

static cl_int err;
static cl_platform_id platform = 0;
static cl_device_id device = 0;

static cl_context_properties props[3] = { CL_CONTEXT_PLATFORM, 0, 0 };
static cl_context ctx = 0;
static cl_command_queue queue = 0;

size_t buf_size = (1 + 2 * MAX_SMI) * MAX_BLOCK_SIZE;
unsigned *res;
static cl_mem dev_res;

static cl_program app;
static cl_kernel kernel;

cl_uint units, freq;


static cl_program compile(const char *source) {
    cl_program app0;
    size_t len = strlen(source);
    
    app0 = clCreateProgramWithSource(ctx, 1, (const char**)&source, &len, &err);
    if (err != CL_SUCCESS) {
	printf("clCreateProgramWithSource() failed with %d\n", err);
	return NULL;
    }
    
    err = clBuildProgram(app0, 1, &device,  "", NULL, NULL);
    if (err != CL_SUCCESS) {
	printf("clBuildProgram() failed with %d\n", err);
	return NULL;
    }

    return app0;
}

int exercise_init(const char *name, size_t size) {
    int ret = 0;
    char param[1024];

    FILE *f;
    size_t len;
    char *source;

    err = clGetPlatformIDs(1, &platform, NULL);
    if (err != CL_SUCCESS) {
        printf( "clGetPlatformIDs() failed with %d\n", err );
        return 1;
    }

    err = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, 1, &device, NULL);
    if (err != CL_SUCCESS) {
        printf( "clGetDeviceIDs() failed with %d\n", err );
        return 1;
    }

    err = clGetDeviceInfo(device, CL_DEVICE_NAME, sizeof(param), param, NULL);
    if (!err) {
	printf("Using device: %s\n", param);
    }


    props[1] = (cl_context_properties)platform;
    ctx = clCreateContext(props, 1, &device, NULL, NULL, &err);
    if (err != CL_SUCCESS) {
        printf( "clCreateContext() failed with %d\n", err );
        return 1;
    }

    queue = clCreateCommandQueue(ctx, device, CL_QUEUE_PROFILING_ENABLE, &err);
    if (err != CL_SUCCESS) {
        printf( "clCreateCommandQueue() failed with %d\n", err );
        clReleaseContext(ctx);
        return 1;
    }

    err = clGetDeviceInfo(device, CL_DEVICE_MAX_COMPUTE_UNITS, sizeof(units), &units, NULL);
    if (err != CL_SUCCESS) {
	printf("Can't determine number of units, error %d\n", err);
	return 1;
    }

    err = clGetDeviceInfo(device, CL_DEVICE_MAX_CLOCK_FREQUENCY, sizeof(freq), &freq, NULL);
    if (err != CL_SUCCESS) {
	printf("Can't determine device frequency, error %d\n", err);
	return 1;
    }


    res = malloc(buf_size * sizeof(unsigned));
    dev_res = clCreateBuffer(ctx, CL_MEM_READ_WRITE, buf_size * sizeof(unsigned), NULL, &err);
    if ((!res)||(err != CL_SUCCESS)) {
	printf("clCreateBuffer() failed with %d\n", err);
	return 1;
    }


    sprintf(param, "%s.cl", name);
    f = fopen(param, "r");
    if (!f) {
	printf("Can't open file with OpenCL kernels\n");
	return 1;
    }

    fseek(f, 0, SEEK_END);
    len = ftell(f);
    fseek(f, 0, SEEK_SET);
    
    source = (char*)malloc(len + 128);
    if (!source) {
	printf("Can't allocate memory for OpenCL source\n");
	return 1;
    }

    sprintf(source, "#define KERNEL_ITERATIONS %lu\n#define MAX_SMI %lu\n\n", KERNEL_ITERATIONS, MAX_SMI);
    if (fread(source + strlen(source), 1, len, f) != len) {
	printf("Can't read OpenCL source\n");
	return 1;
    }
    fclose(f);

    app = compile(source);
    if (!app) return -1;

    free(source);


/*
    kernel = clCreateKernel(app, "pi", &err);
    if (err != CL_SUCCESS) {
	printf("clCreateKernel() failed with %d\n", err);
	return 1;
    }

    clSetKernelArg(kernel, 0, sizeof(cl_mem), &dev_res);
    vector_size = size / (USE_BLOCKS * BLOCK_SIZE);
    clSetKernelArg(kernel, 1, sizeof(unsigned long), &vector_size);
    clSetKernelArg(kernel, 2, sizeof(cl_mem), &scratch);
    clSetKernelArg(kernel, 3, BLOCK_SIZE * sizeof(unsigned long), NULL);
*/

    return 0;
}


void exercise_free() {
    free(res);
    
    clReleaseKernel(kernel);
    clReleaseProgram(app);
    
    clReleaseCommandQueue(queue);
    clReleaseContext(ctx);
    clReleaseMemObject(dev_res);
}

size_t exercise(size_t size, int iterations) {
    int i, j;
    int wrap, simd;
    
    cl_event event;
    cl_event events[iterations];
    struct timeval tv1, tv2;
    unsigned long us;

    size_t runtime = 0;
    cl_ulong start, end, max;

    size_t local_size[] = {MAX_BLOCK_SIZE};
    size_t global_size[] = {MAX_BLOCK_SIZE};

    unsigned *timing;
    unsigned *smi;

	// Cost of running kernel
    kernel = clCreateKernel(app, "empty", &err);
    if (err) {
	printf("clCreateKernel() failed with %d\n", err);
	return -1;
    }

    runtime = 0;
    for (i = 0; i <= iterations; i++) {
	err = clEnqueueNDRangeKernel(queue, kernel, 1, NULL, global_size, local_size, 0, NULL, &event);
	if (err != CL_SUCCESS) {
	    printf("clEnqueueNDRangeKernel() failed with %d\n", err);
	    return -1;
	}
	clFinish(queue);

	if (i) {
	    clGetEventProfilingInfo(event, CL_PROFILING_COMMAND_START, sizeof(cl_ulong), &start, NULL);
	    clGetEventProfilingInfo(event, CL_PROFILING_COMMAND_END, sizeof(cl_ulong), &end, NULL);
	    runtime += end - start;
	}
    }
    printf(" Cost of executing kernel: %.3lf us\n", 1. * runtime / iterations / 1000);


    kernel = clCreateKernel(app, "lanes", &err);
    if (err) {
	printf("clCreateKernel() failed with %d\n", err);
	return -1;
    }
    clSetKernelArg(kernel, 0, sizeof(cl_mem), &dev_res);

    err = clEnqueueNDRangeKernel(queue, kernel, 1, NULL, global_size, local_size, 0, NULL, &event);
    if (err != CL_SUCCESS) {
	printf("clEnqueueNDRangeKernel() failed with %d\n", err);
	return -1;
    }

    err = clEnqueueReadBuffer(queue, dev_res, CL_TRUE, 0, 3 * MAX_BLOCK_SIZE * sizeof(unsigned), res, 0, NULL, NULL);
    if (err != CL_SUCCESS) {
	printf("clEnqueueReadBuffer() failed with %d\n", err);
	return -1;
    }
    
    wrap = 0;
    for (i = 0; i < MAX_BLOCK_SIZE; i++) {
	if (res[i] > wrap) wrap = i;
//	printf("%u %u %u\n", i, res[MAX_BLOCK_SIZE + i], res[2 * MAX_BLOCK_SIZE + i]);
    }
    wrap += 1;

    timing = res + 2 * MAX_BLOCK_SIZE;
    us = timing[0];
    for (i = 1; i < MAX_BLOCK_SIZE; i++) {
	if (timing[i] != us) break;
    }
    simd = i;
	
    printf(" Computational units:  %lu\n", units);
    printf(" Wrap size: %lu\n", wrap);
    printf(" Threads started in parallel: %lu\n", simd);
    printf(" Frquency: %lu MHz\n", freq);


    local_size[0] = simd;
    global_size[0] = 32 * simd;

    kernel = clCreateKernel(app, "shmem_latency", &err);
    if (err) {
	printf("clCreateKernel() failed with %d\n", err);
	return -1;
    }

    clSetKernelArg(kernel, 0, sizeof(cl_mem), &dev_res);
    clSetKernelArg(kernel, 1, KERNEL_ITERATIONS * local_size[0] * sizeof(unsigned), NULL);


    err = clEnqueueNDRangeKernel(queue, kernel, 1, NULL, global_size, local_size, 0, NULL, &event);
    if (err != CL_SUCCESS) {
	printf("clEnqueueNDRangeKernel() failed with %d\n", err);
	return -1;
    }

    err = clEnqueueReadBuffer(queue, dev_res, CL_TRUE, 0, simd * sizeof(unsigned), res, 0, NULL, NULL);
    if (err != CL_SUCCESS) {
	printf("clEnqueueReadBuffer() failed with %d\n", err);
	return -1;
    }

    for (i = 0, j = 0; i < simd; i++) {
	j += res[i];
    }
	

    printf(" Shared memory latency (Bank conflicts): %lu clocks per access\n", j / KERNEL_ITERATIONS / simd);


    local_size[0] = MAX_BLOCK_SIZE;
    global_size[0] = 128 * MAX_BLOCK_SIZE;

    kernel = clCreateKernel(app, "shmem_bandwidth", &err);
    if (err) {
	printf("clCreateKernel() failed with %d\n", err);
	return -1;
    }

    clSetKernelArg(kernel, 0, sizeof(cl_mem), &dev_res);
    clSetKernelArg(kernel, 1, local_size[0] * sizeof(unsigned), NULL);

    for (i = 0; i < units; i++) {
	memset(res + (2 * i    ) * MAX_BLOCK_SIZE, 0x00, MAX_BLOCK_SIZE * sizeof(unsigned));
	memset(res + (2 * i + 1) * MAX_BLOCK_SIZE, 0xFF, MAX_BLOCK_SIZE * sizeof(unsigned));
    }
    err = clEnqueueWriteBuffer(queue, dev_res, CL_TRUE, 0, 2 * units * MAX_BLOCK_SIZE * sizeof(unsigned), res, 0, NULL, NULL);
    if (err != CL_SUCCESS) {
	printf("clEnqueueReadBuffer() failed with %d\n", err);
	return -1;
    }

    err = clEnqueueNDRangeKernel(queue, kernel, 1, NULL, global_size, local_size, 0, NULL, &event);
    if (err != CL_SUCCESS) {
	printf("clEnqueueNDRangeKernel() failed with %d\n", err);
	return -1;
    }

    err = clEnqueueReadBuffer(queue, dev_res, CL_TRUE, 0, 2 * units * MAX_BLOCK_SIZE * sizeof(unsigned), res, 0, NULL, NULL);
    if (err != CL_SUCCESS) {
	printf("clEnqueueReadBuffer() failed with %d\n", err);
	return -1;
    }


    max = 0;
    for (j = 0; j < units; j++) {
        for (i = 0; i < MAX_BLOCK_SIZE; i++) {
	    if (res[(j * 2) * MAX_BLOCK_SIZE + i] > max) max = res[(j * 2) * MAX_BLOCK_SIZE + i];
	}
    }

    printf(" Shared memory bandwidth: %.3lf GB/s (%lf clocks per access)\n", 1. * units * local_size[0] * sizeof(unsigned) * (KERNEL_ITERATIONS - 1) * freq / max / 1024, 1. * max / (KERNEL_ITERATIONS - 1));

    return 0;
}
