#include <sys/types.h>
#include <stdio.h>
#include <string.h>

#include <CL/cl.h>
#include <clAmdBlas.h>


static cl_int err;
static cl_platform_id platform = 0;
static cl_device_id device = 0;

static cl_context_properties props[3] = { CL_CONTEXT_PLATFORM, 0, 0 };
static cl_context ctx = 0;
static cl_command_queue queue = 0;

static cl_mem dev_res, dev_a, dev_b, scratch;

int exercise_required_alignment = 0;

int exercise_init(const char *name, size_t size) {
    int ret = 0;
    char param[1024];

    err = clGetPlatformIDs(1, &platform, NULL);
    if (err != CL_SUCCESS) {
        printf( "clGetPlatformIDs() failed with %d\n", err );
        return 1;
    }

    err = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, 1, &device, NULL);
    if (err != CL_SUCCESS) {
        printf( "clGetDeviceIDs() failed with %d\n", err );
        return 1;
    }

    err = clGetDeviceInfo(device, CL_DEVICE_NAME, sizeof(param), param, NULL);
    if (!err) {
	printf("Using device: %s\n", param);
    }


    props[1] = (cl_context_properties)platform;
    ctx = clCreateContext(props, 1, &device, NULL, NULL, &err);
    if (err != CL_SUCCESS) {
        printf( "clCreateContext() failed with %d\n", err );
        return 1;
    }

    queue = clCreateCommandQueue(ctx, device, CL_QUEUE_PROFILING_ENABLE, &err);
    if (err != CL_SUCCESS) {
        printf( "clCreateCommandQueue() failed with %d\n", err );
        clReleaseContext(ctx);
        return 1;
    }

    err = clAmdBlasSetup();
    if (err != CL_SUCCESS) {
        printf("clAmdBlasSetup() failed with %d\n", err);
        clReleaseCommandQueue(queue);
        clReleaseContext(ctx);
        return 1;
    }

    dev_a = clCreateBuffer(ctx, CL_MEM_READ_ONLY, size * sizeof(float), NULL, &err);
    if (err != CL_SUCCESS) {
	printf("clCreateBuffer() failed with %d\n", err);
	return 1;
    }

    dev_b = clCreateBuffer(ctx, CL_MEM_READ_ONLY, size * sizeof(float), NULL, &err);
    if (err != CL_SUCCESS) {
	printf("clCreateBuffer() failed with %d\n", err);
	return 1;
    }

    dev_res = clCreateBuffer(ctx, CL_MEM_READ_WRITE, sizeof(float), NULL, &err);
    if (err != CL_SUCCESS) {
	printf("clCreateBuffer() failed with %d\n", err);
	return 1;
    }

    scratch = clCreateBuffer(ctx, CL_MEM_READ_WRITE, size * sizeof(float), NULL, &err);
    if (err != CL_SUCCESS) {
	printf("clCreateBuffer() failed with %d\n", err);
	return 1;
    }
    
    return 0;
}

int exercise_allocate(float **a, float **b, size_t size) {
    return  0;
}

void exercise_free() {
    clAmdBlasTeardown();

    clReleaseCommandQueue(queue);
    clReleaseContext(ctx);

    clReleaseMemObject(scratch);
    clReleaseMemObject(dev_res);
    clReleaseMemObject(dev_b);
    clReleaseMemObject(dev_a);
}

size_t exercise(float *res, float *a, float *b, size_t size, int iterations) {
    int i;

    cl_event event = NULL;
    
    struct timeval tv1, tv2;
    unsigned long us = 0;
    
    
    for (i = 0; i < iterations; i++) {
        err = clEnqueueWriteBuffer(queue, dev_a, CL_TRUE, 0, size * sizeof(float), a, 0, NULL, NULL);
	if (err != CL_SUCCESS) {
	    printf("clEnqueueWriteBuffer() failed with %d\n", err);
	    return -1;
	}

	err = clEnqueueWriteBuffer(queue, dev_b, CL_TRUE, 0, size * sizeof(float), b, 0, NULL, NULL);
	if (err != CL_SUCCESS) {
	    printf("clEnqueueWriteBuffer() failed with %d\n", err);
	    return -1;
	}

	    // We need this to avoid some optimizations preventing proper measurements of time		
	err = clEnqueueWriteBuffer(queue, dev_res, CL_TRUE, 0, sizeof(float), res, 0, NULL, &event);
	if (err != CL_SUCCESS) {
	    printf("clEnqueueReadBuffer() failed with %d\n", err);
	    return -1;
	}
	clWaitForEvents(1, &event);

	gettimeofday(&tv1, NULL);
	err = clAmdBlasSdot(size, dev_res, 0, dev_a, 0, 1, dev_b, 0, 1, scratch, 1, &queue, 0, NULL, &event);
	if (err != CL_SUCCESS) {
	    printf("clAmdBlasSdot() failed with %d\n", err);
	    return -1;
	}
	clFinish(queue);
	gettimeofday(&tv2, NULL);

        us += (tv2.tv_sec - tv1.tv_sec)*1000000 + (tv2.tv_usec - tv1.tv_usec);

	err = clEnqueueReadBuffer(queue, dev_res, CL_TRUE, 0, sizeof(float), res, 0, NULL, NULL);
	if (err != CL_SUCCESS) {
	    printf("clEnqueueReadBuffer() failed with %d\n", err);
	    return -1;
	}
    }
	
    return us * 1000;
}
