#include <sys/types.h>
#include <stdio.h>
#include <string.h>

#include <CL/cl.h>

#define BLOCK_SIZE 512
#define USE_BLOCKS 32

#ifndef SHMEM
# define SHMEM 1
#endif

static cl_int err;
static cl_platform_id platform = 0;
static cl_device_id device = 0;

static cl_context_properties props[3] = { CL_CONTEXT_PLATFORM, 0, 0 };
static cl_context ctx = 0;
static cl_command_queue queue = 0;

static cl_mem dev_res, dev_a, dev_b, scratch;

static cl_program app;
static cl_kernel kernel;

static size_t vector_size;

int exercise_required_alignment = BLOCK_SIZE * USE_BLOCKS;

int exercise_init(const char *name, size_t size) {
    int ret = 0;
    char param[1024];

    FILE *f;
    size_t len;
    char *source;

    err = clGetPlatformIDs(1, &platform, NULL);
    if (err != CL_SUCCESS) {
        printf( "clGetPlatformIDs() failed with %d\n", err );
        return 1;
    }

    err = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, 1, &device, NULL);
    if (err != CL_SUCCESS) {
        printf( "clGetDeviceIDs() failed with %d\n", err );
        return 1;
    }

    err = clGetDeviceInfo(device, CL_DEVICE_NAME, sizeof(param), param, NULL);
    if (!err) {
	printf("Using device: %s\n", param);
    }


    props[1] = (cl_context_properties)platform;
    ctx = clCreateContext(props, 1, &device, NULL, NULL, &err);
    if (err != CL_SUCCESS) {
        printf( "clCreateContext() failed with %d\n", err );
        return 1;
    }

    queue = clCreateCommandQueue(ctx, device, CL_QUEUE_PROFILING_ENABLE, &err);
    if (err != CL_SUCCESS) {
        printf( "clCreateCommandQueue() failed with %d\n", err );
        clReleaseContext(ctx);
        return 1;
    }

    dev_a = clCreateBuffer(ctx, CL_MEM_READ_ONLY, size * sizeof(float), NULL, &err);
    if (err != CL_SUCCESS) {
	printf("clCreateBuffer() failed with %d\n", err);
	return 1;
    }

    dev_b = clCreateBuffer(ctx, CL_MEM_READ_ONLY, size * sizeof(float), NULL, &err);
    if (err != CL_SUCCESS) {
	printf("clCreateBuffer() failed with %d\n", err);
	return 1;
    }

    dev_res = clCreateBuffer(ctx, CL_MEM_READ_WRITE, sizeof(float), NULL, &err);
    if (err != CL_SUCCESS) {
	printf("clCreateBuffer() failed with %d\n", err);
	return 1;
    }

    scratch = clCreateBuffer(ctx, CL_MEM_READ_WRITE, size * sizeof(float), NULL, &err);
    if (err != CL_SUCCESS) {
	printf("clCreateBuffer() failed with %d\n", err);
	return 1;
    }

    float res = 0;
    
    err = clEnqueueWriteBuffer(queue, dev_res, CL_TRUE, 0, sizeof(float), &res, 0, NULL, NULL);
    if (err != CL_SUCCESS) {
	printf("clEnqueueWriteBuffer() failed with %d\n", err);
	return -1;
    }

    err = clFinish(queue);
    if (err != CL_SUCCESS) {
	printf("clFinish() failed with %d\n", err);
	return -1;
    }

    sprintf(param, "%s.cl", name);
    f = fopen(param, "r");
    if (!f) {
	printf("Can't open file with OpenCL kernels\n");
	return 1;
    }

    fseek(f, 0, SEEK_END);
    len = ftell(f);
    fseek(f, 0, SEEK_SET);
    
    source = (char*)malloc(len + 128);
    if (!source) {
	printf("Can't allocate memory for OpenCL source\n");
	return 1;
    }
    
    sprintf(source, "#define BLOCK_SIZE %lu\n\n", BLOCK_SIZE);
    if (fread(source + strlen(source), 1, len, f) != len) {
	printf("Can't read OpenCL source\n");
	return 1;
    }
    fclose(f);

    len = strlen(source);

    app = clCreateProgramWithSource(ctx, 1, (const char**)&source, &len, &err);
    if (err != CL_SUCCESS) {
	printf("clCreateProgramWithSource() failed with %d\n", err);
	return 1;
    }
    
    err = clBuildProgram(app, 1, &device,  "", NULL, NULL);
    if (err != CL_SUCCESS) {
	printf("clBuildProgram() failed with %d\n", err);
	return 1;
    }
    
    free(source);
    
    kernel = clCreateKernel(app, "multiply", &err);
    if (err != CL_SUCCESS) {
	printf("clCreateKernel() failed with %d\n", err);
	return 1;
    }

    clSetKernelArg(kernel, 0, sizeof(cl_mem), &dev_res);
    clSetKernelArg(kernel, 1, sizeof(cl_mem), &dev_a);
    clSetKernelArg(kernel, 2, sizeof(cl_mem), &dev_b);
    vector_size = size / (USE_BLOCKS * BLOCK_SIZE);
    clSetKernelArg(kernel, 3, sizeof(unsigned long), &vector_size);
    clSetKernelArg(kernel, 4, sizeof(cl_mem), &scratch);
    clSetKernelArg(kernel, 5, BLOCK_SIZE * sizeof(float), NULL);

    return 0;
}

int exercise_allocate(float **res, float **a, float **b, size_t size) {
    return 0;
}


void exercise_free() {
    clReleaseKernel(kernel);
    clReleaseProgram(app);
    
    clReleaseCommandQueue(queue);
    clReleaseContext(ctx);

    clReleaseMemObject(scratch);
    clReleaseMemObject(dev_res);
    clReleaseMemObject(dev_b);
    clReleaseMemObject(dev_a);
}

size_t exercise(float *res, float *a, float *b, size_t size, int iterations) {
    int i;
    
    cl_event event = NULL;
    struct timeval tv1, tv2;
    unsigned long us = 0;

    cl_ulong start, end;
    cl_ulong runtime = 0;

    size_t local_size[] = {BLOCK_SIZE};
    size_t global_size[] = {USE_BLOCKS * BLOCK_SIZE};

    for (i = 0; i < iterations; i++) {
	err = clEnqueueWriteBuffer(queue, dev_a, CL_TRUE, 0, size * sizeof(float), a, 0, NULL, &event);
        if (err != CL_SUCCESS) {
	    printf("clEnqueueWriteBuffer() failed with %d\n", err);
	    return -1;
	}
	err = clEnqueueWriteBuffer(queue, dev_b, CL_TRUE, 0, size * sizeof(float), b, 0, NULL, &event);
	if (err != CL_SUCCESS) {
	    printf("clEnqueueWriteBuffer() failed with %d\n", err);
	    return -1;
	}

	    // We need this to avoid some optimizations preventing proper measurements of time		
	err = clEnqueueWriteBuffer(queue, dev_res, CL_TRUE, 0, sizeof(float), res, 0, NULL, &event);
	if (err != CL_SUCCESS) {
	    printf("clEnqueueReadBuffer() failed with %d\n", err);
	    return -1;
	}
	clWaitForEvents(1, &event);
	

	gettimeofday(&tv1, NULL);
	err = clEnqueueNDRangeKernel(queue, kernel, 1, NULL, global_size, local_size, 0, NULL, &event);
	if (err != CL_SUCCESS) {
	    printf("clEnqueueNDRangeKernel() failed with %d\n", err);
	    return -1;
	}
	clFinish(queue);
	gettimeofday(&tv2, NULL);
	us += (tv2.tv_sec - tv1.tv_sec)*1000000 + (tv2.tv_usec - tv1.tv_usec);


	clGetEventProfilingInfo(event, CL_PROFILING_COMMAND_START, sizeof(cl_ulong), &start, NULL);
	clGetEventProfilingInfo(event, CL_PROFILING_COMMAND_END, sizeof(cl_ulong), &end, NULL);
	runtime += end - start;

	err = clEnqueueReadBuffer(queue, dev_res, CL_TRUE, 0, sizeof(float), res, 0, NULL, NULL);
	if (err != CL_SUCCESS) {
	    printf("clEnqueueReadBuffer() failed with %d\n", err);
	    return -1;
	}
    }

//    return us * 1000;
    return runtime;
}

