#include <sys/types.h>
#include <stdio.h>
#include <string.h>

#include <CL/cl.h>
#include <clAmdBlas.h>


#ifndef PARALLEL
# define PARALLEL 1
#endif

static cl_int err;
static cl_platform_id platform = 0;
static cl_device_id device = 0;

static cl_context_properties props[3] = { CL_CONTEXT_PLATFORM, 0, 0 };
static cl_context ctx = 0;
static cl_command_queue queue, queues[PARALLEL] = {0};

static cl_mem dev_res, dev_a, dev_b;

int exercise_required_alignment = 0;

int exercise_init(const char *name, size_t size) {
    int i;
    int ret = 0;
    char param[1024];

    err = clGetPlatformIDs(1, &platform, NULL);
    if (err != CL_SUCCESS) {
        printf( "clGetPlatformIDs() failed with %d\n", err );
        return 1;
    }

    err = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, 1, &device, NULL);
    if (err != CL_SUCCESS) {
        printf( "clGetDeviceIDs() failed with %d\n", err );
        return 1;
    }

    err = clGetDeviceInfo(device, CL_DEVICE_NAME, sizeof(param), param, NULL);
    if (!err) {
	printf("Using device: %s\n", param);
    }


    props[1] = (cl_context_properties)platform;
    ctx = clCreateContext(props, 1, &device, NULL, NULL, &err);
    if (err != CL_SUCCESS) {
        printf( "clCreateContext() failed with %d\n", err );
        return 1;
    }

    for (i = 0; i < PARALLEL; i++) {
	queues[i] = clCreateCommandQueue(ctx, device, CL_QUEUE_PROFILING_ENABLE, &err);
	if (err != CL_SUCCESS) {
    	    printf( "clCreateCommandQueue() failed with %d\n", err );
    	    clReleaseContext(ctx);
	    return 1;
	}
    }
    queue = queues[0];

    err = clAmdBlasSetup();
    if (err != CL_SUCCESS) {
        printf("clAmdBlasSetup() failed with %d\n", err);
        clReleaseCommandQueue(queue);
        clReleaseContext(ctx);
        return 1;
    }

    dev_a = clCreateBuffer(ctx, CL_MEM_READ_ONLY, size * size * sizeof(float), NULL, &err);
    if (err != CL_SUCCESS) {
	printf("clCreateBuffer() failed with %d\n", err);
	return 1;
    }

    dev_b = clCreateBuffer(ctx, CL_MEM_READ_ONLY, size * size * sizeof(float), NULL, &err);
    if (err != CL_SUCCESS) {
	printf("clCreateBuffer() failed with %d\n", err);
	return 1;
    }

    dev_res = clCreateBuffer(ctx, CL_MEM_READ_WRITE, size * size * sizeof(float), NULL, &err);
    if (err != CL_SUCCESS) {
	printf("clCreateBuffer() failed with %d\n", err);
	return 1;
    }
    
    return 0;
}

int exercise_allocate(float **res, float **a, float **b, size_t size) {
    return  0;
}

void exercise_free() {
    int i;
    clAmdBlasTeardown();

    for (i = 0; i < PARALLEL; i++) {
	clReleaseCommandQueue(queues[i]);
    }
    clReleaseContext(ctx);

    clReleaseMemObject(dev_res);
    clReleaseMemObject(dev_b);
    clReleaseMemObject(dev_a);
}

size_t exercise(float *res, float *a, float *b, size_t size, int iterations) {
    int i,j;

    cl_event event = NULL;
    size_t runtime = 0;
    cl_ulong start, end;
    
    float alpha = 1, beta = 0;
    static const clAmdBlasTranspose transA = clAmdBlasNoTrans, transB = clAmdBlasNoTrans;
    static const clAmdBlasOrder order = clAmdBlasRowMajor;

    for (i = 0; i < iterations; i++) {
	err = clEnqueueWriteBuffer(queue, dev_a, CL_TRUE, 0, size * size * sizeof(float), a, 0, NULL, NULL);
	if (err != CL_SUCCESS) {
	    printf("clEnqueueWriteBuffer() failed with %d\n", err);
	    return -1;
	}

	err = clEnqueueWriteBuffer(queue, dev_b, CL_TRUE, 0, size * size * sizeof(float), b, 0, NULL, NULL);
	if (err != CL_SUCCESS) {
	    printf("clEnqueueWriteBuffer() failed with %d\n", err);
	    return -1;
	}

	    // Memory corruption above 2 with current version of clBlas, i is set to 0
	err = clAmdBlasSgemm(order, transA, transB, size, size, size, alpha, dev_a, size, dev_b, size, beta, dev_res, size, PARALLEL, queues, 0, NULL, &event);
	if (err != CL_SUCCESS) {
	    printf("clAmdBlasSgemm() failed with %d\n", err);
	    return -1;
	}

	for (j = 0; j < PARALLEL; j++) {
	    err = clFinish(queues[j]);
	    if (err != CL_SUCCESS) {
		printf("clFinish() failed with %d\n", err);
		return -1;
	    }
	}

	err = clEnqueueReadBuffer(queue, dev_res, CL_TRUE, 0, size * size * sizeof(float), res, 0, NULL, NULL);
	if (err != CL_SUCCESS) {
	    printf("clEnqueueReadBuffer() failed with %d\n", err);
	    return -1;
	}
	
	err = clGetEventProfilingInfo(event, CL_PROFILING_COMMAND_START, sizeof(cl_ulong), &start, NULL);
	if (err != CL_SUCCESS) {
	    printf("clGetEventProfilingInfo() failed with %d\n", err);
	    return -1;
	}
	
	err = clGetEventProfilingInfo(event, CL_PROFILING_COMMAND_END, sizeof(cl_ulong), &end, NULL);
	if (err != CL_SUCCESS) {
	    printf("clGetEventProfilingInfo() failed with %d\n", err);
	    return -1;
	}
	
	runtime += end - start;
    }

    return runtime;
}
