#include <sys/types.h>
#include <stdio.h>
#include <string.h>

#include <CL/cl.h>

#define BLOCK_SIZE 16

#ifndef PPT
# define PPT 1
#endif 

#ifndef SHMEM
# define SHMEM 1
#endif

static cl_int err;
static cl_platform_id platform = 0;
static cl_device_id device = 0;

static cl_context_properties props[3] = { CL_CONTEXT_PLATFORM, 0, 0 };
static cl_context ctx = 0;
static cl_command_queue queues[2] = {0};
static cl_command_queue queue = 0;

static cl_mem dev_res[2], dev_a[2], dev_b[2];
static cl_mem pinned_mem_res, pinned_mem_a, pinned_mem_b;
static float *pinned_res, *pinned_a, *pinned_b;

static cl_program app;
static cl_kernel kernel;

static size_t matrix_size;

int exercise_required_alignment = PPT * BLOCK_SIZE;

int exercise_init(const char *name, size_t size) {
    int ret = 0;
    char param[1024];

    FILE *f;
    size_t len;
    char *source;
    
    int i;

    err = clGetPlatformIDs(1, &platform, NULL);
    if (err != CL_SUCCESS) {
        printf( "clGetPlatformIDs() failed with %d\n", err );
        return 1;
    }

    err = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, 1, &device, NULL);
    if (err != CL_SUCCESS) {
        printf( "clGetDeviceIDs() failed with %d\n", err );
        return 1;
    }

    err = clGetDeviceInfo(device, CL_DEVICE_NAME, sizeof(param), param, NULL);
    if (!err) {
	printf("Using device: %s\n", param);
    }


    props[1] = (cl_context_properties)platform;
    ctx = clCreateContext(props, 1, &device, NULL, NULL, &err);
    if (err != CL_SUCCESS) {
        printf( "clCreateContext() failed with %d\n", err );
        return 1;
    }

    for (i = 0; i < 2; i++) {
        queues[i] = clCreateCommandQueue(ctx, device, CL_QUEUE_PROFILING_ENABLE, &err);
	if (err != CL_SUCCESS) {
    	    printf( "clCreateCommandQueue() failed with %d\n", err );
    	    clReleaseContext(ctx);
    	    return 1;
	}

	dev_a[i] = clCreateBuffer(ctx, CL_MEM_READ_ONLY, size * size * sizeof(float), NULL, &err);
	if (err != CL_SUCCESS) {
	    printf("clCreateBuffer() failed with %d\n", err);
	    return 1;
	}

	dev_b[i] = clCreateBuffer(ctx, CL_MEM_READ_ONLY, size * size * sizeof(float), NULL, &err);
	if (err != CL_SUCCESS) {
	    printf("clCreateBuffer() failed with %d\n", err);
	    return 1;
	}

	dev_res[i] = clCreateBuffer(ctx, CL_MEM_READ_WRITE, size * size * sizeof(float), NULL, &err);
	if (err != CL_SUCCESS) {
	    printf("clCreateBuffer() failed with %d\n", err);
	    return 1;
	}
    }
    queue = queues[0];

    char *res = calloc(size * size, sizeof(float));
    if (res) {
	for (i = 0; i < 2; i++) {
	    err = clEnqueueWriteBuffer(queue, dev_res[i], CL_TRUE, 0, size * size * sizeof(float), res, 0, NULL, NULL);
	    if (err != CL_SUCCESS) {
		printf("clEnqueueWriteBuffer() failed with %d\n", err);
		return -1;
	    }
	}
	free(res);

        err = clFinish(queue);
	if (err != CL_SUCCESS) {
	    printf("clFinish() failed with %d\n", err);
	    return -1;
	}
    }

    sprintf(param, "%s.cl", name);
    f = fopen(param, "r");
    if (!f) {
	printf("Can't open file with OpenCL kernels\n");
	return 1;
    }

    fseek(f, 0, SEEK_END);
    len = ftell(f);
    fseek(f, 0, SEEK_SET);
    
    source = (char*)malloc(len + 128);
    if (!source) {
	printf("Can't allocate memory for OpenCL source\n");
	return 1;
    }
    
    sprintf(source, "#define BLOCK_SIZE %lu\n#define PPT %lu\n\n", BLOCK_SIZE, PPT);
    if (fread(source + strlen(source), 1, len, f) != len) {
	printf("Can't read OpenCL source\n");
	return 1;
    }
    fclose(f);

    len = strlen(source);

    app = clCreateProgramWithSource(ctx, 1, (const char**)&source, &len, &err);
    if (err != CL_SUCCESS) {
	printf("clCreateProgramWithSource() failed with %d\n", err);
	return 1;
    }
    
    err = clBuildProgram(app, 1, &device,  "", NULL, NULL);
    if (err != CL_SUCCESS) {
	printf("clBuildProgram() failed with %d\n", err);
	return 1;
    }
    
    free(source);
    
    kernel = clCreateKernel(app, "multiply", &err);
    if (err != CL_SUCCESS) {
	printf("clCreateKernel() failed with %d\n", err);
	return 1;
    }

    matrix_size = size;
    clSetKernelArg(kernel, 3, sizeof(unsigned long), &matrix_size);
    err = clSetKernelArg(kernel, 4, 2 * SHMEM * BLOCK_SIZE * BLOCK_SIZE * sizeof(float), NULL);

    return 0;
}

int exercise_allocate(float **res, float **a, float **b, size_t size) {
    pinned_mem_a = clCreateBuffer(ctx, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, size * size * sizeof(float), NULL, &err);
    if (err != CL_SUCCESS) {
	printf("clCreateBuffer() for pinned memory failed with %d\n", err);
	return 1;
    }
    
    pinned_a = (cl_float*)clEnqueueMapBuffer(queue, pinned_mem_a, CL_TRUE, CL_MAP_WRITE, 0, size * size * sizeof(float), 0, NULL, NULL, &err);
    if (err != CL_SUCCESS) {
	printf("clEnqueueMapBuffer() failed with %d\n", err);
	return 1;
    }

    pinned_mem_b = clCreateBuffer(ctx, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, size * size * sizeof(float), NULL, &err);
    if (err != CL_SUCCESS) {
	printf("clCreateBuffer() for pinned memory failed with %d\n", err);
	return 1;
    }
    
    pinned_b = (cl_float*)clEnqueueMapBuffer(queue, pinned_mem_b, CL_TRUE, CL_MAP_WRITE, 0, size * size * sizeof(float), 0, NULL, NULL, &err);
    if (err != CL_SUCCESS) {
	printf("clEnqueueMapBuffer() failed with %d\n", err);
	return 1;
    }

    pinned_mem_res = clCreateBuffer(ctx, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, size * size * sizeof(float), NULL, &err);
    if (err != CL_SUCCESS) {
	printf("clCreateBuffer() for pinned memory failed with %d\n", err);
	return 1;
    }
    
    pinned_res = (cl_float*)clEnqueueMapBuffer(queue, pinned_mem_res, CL_TRUE, CL_MAP_READ, 0, size * size * sizeof(float), 0, NULL, NULL, &err);
    if (err != CL_SUCCESS) {
	printf("clEnqueueMapBuffer() failed with %d\n", err);
	return 1;
    }
    
    *res = pinned_res;
    *a = pinned_a;
    *b = pinned_b;

    return 0;
}


void exercise_free() {
    int i;
    
    clReleaseKernel(kernel);
    clReleaseProgram(app);

    clEnqueueUnmapMemObject(queue, pinned_mem_a, pinned_a, 0, NULL, NULL);
    clEnqueueUnmapMemObject(queue, pinned_mem_b, pinned_b, 0, NULL, NULL);
    clEnqueueUnmapMemObject(queue, pinned_mem_res, pinned_res, 0, NULL, NULL);

    clReleaseMemObject(pinned_mem_a);
    clReleaseMemObject(pinned_mem_b);
    clReleaseMemObject(pinned_mem_res);

    for (i = 0; i < 2; i++) {
	clReleaseCommandQueue(queues[i]);
	clReleaseMemObject(dev_res[i]);
	clReleaseMemObject(dev_b[i]);
	clReleaseMemObject(dev_a[i]);
    }

    clReleaseContext(ctx);
}

size_t exercise(float *res, float *a, float *b, size_t size, int iterations) {
    int i, pos;
    
    cl_event events[iterations];
    size_t runtime = 0;
    cl_ulong start, end;

    size_t local_size[] = {BLOCK_SIZE, BLOCK_SIZE};
    size_t global_size[] = {size/PPT, size/PPT};

    for (i = 0; i <= iterations; i++) {
	if (i < iterations) {
	    pos = i%2;
	    
	    err = clEnqueueWriteBuffer(queues[pos], dev_a[pos], CL_FALSE, 0, size * size * sizeof(float), a, 0, NULL, NULL);
	    if (err != CL_SUCCESS) {
		printf("clEnqueueWriteBuffer() failed with %d\n", err);
		return -11;
	    }

	    err = clEnqueueWriteBuffer(queues[pos], dev_b[pos], CL_FALSE, 0, size * size * sizeof(float), b, 0, NULL, NULL);
	    if (err != CL_SUCCESS) {
		printf("clEnqueueWriteBuffer() failed with %d\n", err);
		return -1;
	    }
	    
	    if (!i) continue;
	}
	
	pos = (i + 1)%2;

	clSetKernelArg(kernel, 0, sizeof(cl_mem), &dev_res[pos]);
	clSetKernelArg(kernel, 1, sizeof(cl_mem), &dev_a[pos]);
	clSetKernelArg(kernel, 2, sizeof(cl_mem), &dev_b[pos]);

	err = clEnqueueNDRangeKernel(queues[pos], kernel, 2, 0, global_size, local_size, 0, NULL, &events[i-1]);
	if (err != CL_SUCCESS) {
	    printf("clEnqueueNDRangeKernel() failed with %d\n", err);
	    return -1;
	}

	err = clEnqueueReadBuffer(queues[pos], dev_res[pos], CL_FALSE, 0, size * size * sizeof(float), res, 0, NULL, NULL);
	if (err != CL_SUCCESS) {
	    printf("clEnqueueReadBuffer() failed with %d\n", err);
	    return -1;
	}
    }

    err = clFinish(queues[0]);
    if (err != CL_SUCCESS) {
	printf("clFinish() failed with %d\n", err);
	return -1;
    }

    err = clFinish(queues[1]);
    if (err != CL_SUCCESS) {
	printf("clFinish() failed with %d\n", err);
	return -1;
    }

	
    for (i = 0; i < iterations; i++) {
	err = clGetEventProfilingInfo(events[i], CL_PROFILING_COMMAND_START, sizeof(cl_ulong), &start, NULL);
	if (err != CL_SUCCESS) {
	    printf("clGetEventProfilingInfo() failed with %d\n", err);
	    return -1;
	}
	
	err = clGetEventProfilingInfo(events[i], CL_PROFILING_COMMAND_END, sizeof(cl_ulong), &end, NULL);
	if (err != CL_SUCCESS) {
	    printf("clGetEventProfilingInfo() failed with %d\n", err);
	    return -1;
	}
	
	runtime += end - start;
    }


    return runtime;
}

