#include <sys/types.h>
#include <stdio.h>
#include <string.h>
#include <pthread.h>

#define HAVE_SSE

#ifdef HAVE_SSE
# include <xmmintrin.h>
#endif /* HAVE_SSE */

#include <CL/cl.h>

#define BLOCK_SIZE 16

#ifndef PPT
# define PPT 1
#endif 

#ifndef SHMEM
# define SHMEM 1
#endif

#ifndef PARALLEL
# define PARALLEL 8
#endif

#define BASE_SIZE (BLOCK_SIZE * BLOCK_SIZE * PPT)

//#define DEBUG

static cl_int err;
static cl_platform_id platform = 0;

static cl_context_properties props[3] = { CL_CONTEXT_PLATFORM, 0, 0 };
static cl_context ctx = 0;
static cl_command_queue queue = 0;
static cl_command_queue queues[PARALLEL] = {0};
static cl_program app;
static cl_kernel kernel[PARALLEL];

cl_mem dev_res[PARALLEL], dev_a[PARALLEL], dev_b[PARALLEL];
static cl_mem pinned_mem_res = NULL, pinned_mem_a = NULL, pinned_mem_b = NULL;
static float *pinned_res = NULL, *pinned_a = NULL, *pinned_b = NULL;
static float *tmp_res[PARALLEL];

static pthread_mutex_t blocks_mutex = PTHREAD_MUTEX_INITIALIZER;
static pthread_mutex_t *res_mutexes;

typedef struct {
    size_t ijk, pos, size;
    float *a, *b, *res;
    cl_event event;
    int stopped;
} exercise_block_info_t;

size_t current_block = 0;
unsigned long base_size = BASE_SIZE;
unsigned long last_block_size;
size_t allocated_blocks, blocks;
static exercise_block_info_t block_info[PARALLEL] = {0};

static size_t runtime = 0;

int exercise_required_alignment = 0;

int exercise_init(const char *name, size_t size) {
    int ret = 0;
    char param[1024];

    FILE *f;
    size_t len;
    char *source;

    int i;
    
    cl_uint num_devices;
    cl_device_id devices[PARALLEL] = {0};


    allocated_blocks = size / BASE_SIZE;
    if (allocated_blocks * BASE_SIZE < size) ++allocated_blocks;

    res_mutexes = (pthread_mutex_t*)malloc(allocated_blocks * allocated_blocks * sizeof(pthread_mutex_t));
    if (!res_mutexes) {
	printf("Memory allocation failed\n");
	return 1;
    }
    
    for (i = 0; i < allocated_blocks * allocated_blocks; i++) {
	pthread_mutex_init(res_mutexes + i, NULL);
    }

    size = BASE_SIZE;

    err = clGetPlatformIDs(1, &platform, NULL);
    if (err != CL_SUCCESS) {
        printf( "clGetPlatformIDs() failed with %d\n", err );
        return 1;
    }

    err = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, PARALLEL, devices, &num_devices);
    if (err != CL_SUCCESS) {
        printf( "clGetDeviceIDs() failed with %d\n", err );
        return 1;
    }

    printf("Found %u devices: ", num_devices);
    for (i = 0; i < num_devices; i++) {
	err = clGetDeviceInfo(devices[i], CL_DEVICE_NAME, sizeof(param), param, NULL);
	if (i) printf(", ");
	if (err) printf("???");
	else printf("%s", param);
    }
    printf("\n");

    props[1] = (cl_context_properties)platform;
    ctx = clCreateContext(props, num_devices, devices, NULL, NULL, &err);
    if (err != CL_SUCCESS) {
        printf( "clCreateContext() failed with %d\n", err );
        return 1;
    }

    for (i = 0; i < PARALLEL; i++) {
	tmp_res[i] = (float*)malloc(size * size * sizeof(float));
	if (!tmp_res[i]) {
	    printf("Can't allocate temporary memory buffer\n");
	    return 1;
	}

        queues[i] = clCreateCommandQueue(ctx, devices[i%num_devices], CL_QUEUE_PROFILING_ENABLE, &err);
	if (err != CL_SUCCESS) {
    	    printf( "clCreateCommandQueue() failed with %d\n", err );
    	    clReleaseContext(ctx);
    	    return 1;
	}

	dev_a[i] = clCreateBuffer(ctx, CL_MEM_READ_ONLY, size * size * sizeof(float), NULL, &err);
	if (err != CL_SUCCESS) {
	    printf("clCreateBuffer() failed with %d\n", err);
	    return 1;
	}

	dev_b[i] = clCreateBuffer(ctx, CL_MEM_READ_ONLY, size * size * sizeof(float), NULL, &err);
	if (err != CL_SUCCESS) {
	    printf("clCreateBuffer() failed with %d\n", err);
	    return 1;
	}

	dev_res[i] = clCreateBuffer(ctx, CL_MEM_READ_WRITE, size * size * sizeof(float), NULL, &err);
	if (err != CL_SUCCESS) {
	    printf("clCreateBuffer() failed with %d\n", err);
	    return 1;
	}
    }
    queue = queues[0];

    char *res = calloc(size * size, sizeof(float));
    if (res) {
	for (i = 0; i < PARALLEL; i++) {
	    err = clEnqueueWriteBuffer(queues[i], dev_res[i], CL_TRUE, 0, size * size * sizeof(float), res, 0, NULL, NULL);
	    if (err != CL_SUCCESS) {
		printf("clEnqueueWriteBuffer() failed with %d\n", err);
		return -1;
	    }

	    err = clEnqueueWriteBuffer(queues[i], dev_a[i], CL_TRUE, 0, size * size * sizeof(float), res, 0, NULL, NULL);
	    if (err != CL_SUCCESS) {
		printf("clEnqueueWriteBuffer() failed with %d\n", err);
		return -1;
	    }

	    err = clEnqueueWriteBuffer(queues[i], dev_b[i], CL_TRUE, 0, size * size * sizeof(float), res, 0, NULL, NULL);
	    if (err != CL_SUCCESS) {
		printf("clEnqueueWriteBuffer() failed with %d\n", err);
		return -1;
	    }
	}
	free(res);

        err = clFinish(queue);
	if (err != CL_SUCCESS) {
	    printf("clFinish() failed with %d\n", err);
	    return -1;
	}
    }

    sprintf(param, "%s.cl", name);
    f = fopen(param, "r");
    if (!f) {
	printf("Can't open file with OpenCL kernels\n");
	return 1;
    }

    fseek(f, 0, SEEK_END);
    len = ftell(f);
    fseek(f, 0, SEEK_SET);
    
    source = (char*)malloc(len + 128);
    if (!source) {
	printf("Can't allocate memory for OpenCL source\n");
	return 1;
    }
    
    sprintf(source, "#define BLOCK_SIZE %lu\n#define PPT %lu\n#define pitch %lu\n\n", BLOCK_SIZE, PPT, BASE_SIZE);
    if (fread(source + strlen(source), 1, len, f) != len) {
	printf("Can't read OpenCL source\n");
	return 1;
    }
    fclose(f);

    len = strlen(source);

    app = clCreateProgramWithSource(ctx, 1, (const char**)&source, &len, &err);
    if (err != CL_SUCCESS) {
	printf("clCreateProgramWithSource() failed with %d\n", err);
	return 1;
    }
    
    err = clBuildProgram(app, num_devices, devices,  "", NULL, NULL);
    if (err != CL_SUCCESS) {
	printf("clBuildProgram() failed with %d\n", err);
	return 1;
    }
    
    free(source);
    
    for (i = 0; i < PARALLEL; i++) {
        kernel[i] = clCreateKernel(app, "multiply", &err);
	if (err != CL_SUCCESS) {
	    printf("clCreateKernel() failed with %d\n", err);
	    return 1;
	}

	clSetKernelArg(kernel[i], 0, sizeof(cl_mem), &dev_res[i]);
	clSetKernelArg(kernel[i], 1, sizeof(cl_mem), &dev_a[i]);
	clSetKernelArg(kernel[i], 2, sizeof(cl_mem), &dev_b[i]);
	clSetKernelArg(kernel[i], 4, 2 * SHMEM * BLOCK_SIZE * BLOCK_SIZE * sizeof(float), NULL);
    }

    return 0;
}

static void exercise_clean_pinned() {
    if (pinned_a) clEnqueueUnmapMemObject(queue, pinned_mem_a, pinned_a, 0, NULL, NULL);
    if (pinned_b) clEnqueueUnmapMemObject(queue, pinned_mem_b, pinned_b, 0, NULL, NULL);
    if (pinned_res) clEnqueueUnmapMemObject(queue, pinned_mem_res, pinned_res, 0, NULL, NULL);

    if (pinned_mem_a) clReleaseMemObject(pinned_mem_a);
    if (pinned_mem_b) clReleaseMemObject(pinned_mem_b);
    if (pinned_mem_res) clReleaseMemObject(pinned_mem_res);
    
    pinned_a = pinned_b = pinned_res = NULL;
    pinned_mem_a = pinned_mem_b = pinned_mem_res = NULL;
}

int exercise_allocate(float **res, float **a, float **b, size_t size) {
    pinned_mem_a = clCreateBuffer(ctx, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, size * size * sizeof(float), NULL, &err);
    if (err != CL_SUCCESS) {
	printf("clCreateBuffer() for pinned memory failed with %d\n", err);
	exercise_clean_pinned();
	return 1;
    }
    
    pinned_a = (cl_float*)clEnqueueMapBuffer(queue, pinned_mem_a, CL_TRUE, CL_MAP_WRITE, 0, size * size * sizeof(float), 0, NULL, NULL, &err);
    if (err != CL_SUCCESS) {
	printf("clEnqueueMapBuffer() failed with %d\n", err);
	exercise_clean_pinned();
	return 1;
    }

    pinned_mem_b = clCreateBuffer(ctx, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, size * size * sizeof(float), NULL, &err);
    if (err != CL_SUCCESS) {
	printf("clCreateBuffer() for pinned memory failed with %d\n", err);
	exercise_clean_pinned();
	return 1;
    }
    
    pinned_b = (cl_float*)clEnqueueMapBuffer(queue, pinned_mem_b, CL_TRUE, CL_MAP_WRITE, 0, size * size * sizeof(float), 0, NULL, NULL, &err);
    if (err != CL_SUCCESS) {
	printf("clEnqueueMapBuffer() failed with %d\n", err);
	exercise_clean_pinned();
	return 1;
    }

    pinned_mem_res = clCreateBuffer(ctx, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, size * size * sizeof(float), NULL, &err);
    if (err != CL_SUCCESS) {
	printf("clCreateBuffer() for pinned memory failed with %d\n", err);
	exercise_clean_pinned();
	return 1;
    }
    
    pinned_res = (cl_float*)clEnqueueMapBuffer(queue, pinned_mem_res, CL_TRUE, CL_MAP_READ, 0, size * size * sizeof(float), 0, NULL, NULL, &err);
    if (err != CL_SUCCESS) {
	printf("clEnqueueMapBuffer() failed with %d\n", err);
	exercise_clean_pinned();
	return 1;
    }


    *res = pinned_res;
    *a = pinned_a;
    *b = pinned_b;

    return 0;
}



void exercise_free() {
    int i;

    exercise_clean_pinned();    

    for (i = 0; i < PARALLEL; i++) {
	clReleaseKernel(kernel[i]);
	clReleaseCommandQueue(queues[i]);
	clReleaseMemObject(dev_res[i]);
	clReleaseMemObject(dev_b[i]);
	clReleaseMemObject(dev_a[i]);
	free(tmp_res[i]);
    }

    clReleaseProgram(app);
    clReleaseContext(ctx);

    for (i = 0; i < allocated_blocks * allocated_blocks; i++) {
	pthread_mutex_destroy(res_mutexes + i);
    }
    free(res_mutexes);

}


static cl_event exercise_schedule_block(exercise_block_info_t *info);

static void exercise_block_finished_cb(cl_event event, cl_int event_command_exec_status, void *user_data) {
    cl_int err;
    cl_event ready = NULL;
    size_t pos;
    size_t i, j, ij, ijk;
    size_t x, y;
    size_t real_size_x, real_size_y;

    cl_ulong start, end;

    exercise_block_info_t *info = (exercise_block_info_t*)user_data;

    ijk = info->ijk;

    if (event_command_exec_status != CL_COMPLETE) {
	printf("A failure is registered by event callback\n");
	info->stopped = 1;
	return;
    }

    if (runtime != (size_t)-2) {
	err = clGetEventProfilingInfo(info->event, CL_PROFILING_COMMAND_START, sizeof(cl_ulong), &start, NULL);
	if (err != CL_SUCCESS) {
	    printf("clGetEventProfilingInfo() failed with %d\n", err);
	    runtime = (size_t)-2;
//	    info->stopped = 1;
//	    return;
	}
    }
    
    if (runtime != (size_t)-2) {
	err = clGetEventProfilingInfo(info->event, CL_PROFILING_COMMAND_END, sizeof(cl_ulong), &end, NULL);
	if (err != CL_SUCCESS) {
	    printf("clGetEventProfilingInfo() failed with %d\n", err);
	    runtime = (size_t)-2;
//	    info->stopped = 1;
//	    return;
	}
    }
    
    if (runtime != (size_t)-2) runtime += end - start;

    ready = exercise_schedule_block(info);

    ij = ijk % (blocks * blocks);
    i = ij / blocks; j = ij % blocks;

    if ((i + 1) == blocks) real_size_y = info->size - (blocks - 1) * BASE_SIZE;
    else real_size_y = BASE_SIZE;

    if ((j + 1) == blocks) real_size_x = info->size - (blocks - 1) * BASE_SIZE;
    else real_size_x = BASE_SIZE;

    err = pthread_mutex_lock(res_mutexes + ij);
    if (err) {
	printf("pthread_mutex_lock() failed with %d\n", err);
	info->stopped = 1;
	return;
    }
    
#ifdef HAVE_SSE
    if ((info->size%4) == 0) {
	for (y = 0; y < real_size_y; ++y) {
	    __m128 *dst = (__m128*)(info->res + (i * BASE_SIZE + y) * info->size + (j * BASE_SIZE));
	    __m128 *src = (__m128*)(tmp_res[info->pos] + y * real_size_x);

	    for (x = 0; x < real_size_x/4; ++x) {
		dst[x] = _mm_add_ps(dst[x], src[x]);
	    }
	}
    } else 
#endif
    {
	for (y = 0; y < real_size_y; ++y) {
	    for (x = 0; x < real_size_x; ++x) {
		info->res[(i * BASE_SIZE + y) * info->size + (j * BASE_SIZE + x)] += tmp_res[info->pos][y * real_size_x + x];
	    }
	}
    }

    pthread_mutex_unlock(res_mutexes + ij);

#ifdef DEBUG    
    printf(" Finished step %lu of block %lux%lu, queue %lu\n", ijk / (blocks * blocks), i, j, info->pos);
#endif

    if (ready) {
	clSetUserEventStatus(ready, CL_COMPLETE);
	clReleaseEvent(ready);
    } else {
#ifdef DEBUG
    	printf(" Terminating queue %lu\n", info->pos);
#endif
	info->stopped = 1;
    }
}

static cl_event exercise_schedule_block(exercise_block_info_t *info) {
    cl_event cbevent, ready;
    size_t real_size, real_size_x, real_size_y;
    size_t i, j, k, ij;
    size_t pos = info->pos;
    unsigned long *kernel_size;

    err = pthread_mutex_lock(&blocks_mutex);
    if (err) {
	printf("pthread_mutex_lock() failed with %d\n", err);
	return NULL;
    }
    if (current_block < blocks * blocks * blocks) {
	info->ijk = current_block++;
    } else {
	pthread_mutex_unlock(&blocks_mutex);
	return NULL;
    }
    pthread_mutex_unlock(&blocks_mutex);

    ready = clCreateUserEvent(ctx, &err);
    if (err != CL_SUCCESS) {
	printf("clCreateUserEvent() failed with %d\n", err);
	return NULL;
    }

    k = info->ijk / (blocks * blocks); ij = info->ijk % (blocks * blocks);
    i = ij / blocks; j = ij % blocks;

    if ((k + 1) == blocks) {
	real_size = last_block_size;
	kernel_size = &last_block_size;
    } else {
	real_size = BASE_SIZE;
	kernel_size = &base_size;
    }

    if ((i + 1) == blocks) real_size_y = last_block_size;
    else real_size_y = BASE_SIZE;

    if ((j + 1) == blocks) real_size_x = last_block_size;
    else real_size_x = BASE_SIZE;

#ifdef DEBUG
    printf(" Scheduling step %lu of block %lux%lu to queue %lu, size %lu\n", k, i, j, info->pos, real_size);
#endif

    size_t padded_size = get_padded_size(real_size, PPT * BLOCK_SIZE);
    size_t padded_size_y = get_padded_size(real_size_y, PPT * BLOCK_SIZE);
    size_t padded_size_x = get_padded_size(real_size_x, PPT * BLOCK_SIZE);
    
    size_t global_size[] = {padded_size_x/PPT, padded_size_y/PPT};
    size_t local_size[] = {BLOCK_SIZE, BLOCK_SIZE};

    const size_t origin[3] = {0,0,0};
    const size_t region_a[3] = {real_size * sizeof(float), real_size_y, 1};
    const size_t region_b[3] = {real_size_x * sizeof(float), real_size, 1};
    const size_t region_out[3] = {real_size_x * sizeof(float), real_size_y, 1};

    const size_t origin_a[3] = {k * BASE_SIZE * sizeof(float), i * BASE_SIZE, 0};
    const size_t origin_b[3] = {j * BASE_SIZE * sizeof(float), k * BASE_SIZE, 0};

    err = clEnqueueWriteBufferRect(queues[pos], dev_a[pos], CL_FALSE, origin, origin_a, region_a, BASE_SIZE * sizeof(float), 0, info->size * sizeof(float), 0,  info->a, 0, NULL, NULL);
    if (err != CL_SUCCESS) {
	printf("clEnqueueWriteBufferRect() failed with %d\n", err);
	return NULL;
    }

    err = clEnqueueWriteBufferRect(queues[pos], dev_b[pos], CL_FALSE, origin, origin_b, region_b, BASE_SIZE * sizeof(float), 0, info->size * sizeof(float), 0,  info->b, 0, NULL, NULL);
    if (err != CL_SUCCESS) {
	printf("clEnqueueWriteBufferRect() failed with %d\n", err);
	return NULL;
    }

    clSetKernelArg(kernel[pos], 3, sizeof(unsigned long), kernel_size);

    err = clEnqueueNDRangeKernel(queues[pos], kernel[pos], 2, 0, global_size, local_size, 0, NULL, &info->event);
    if (err != CL_SUCCESS) {
	printf("clEnqueueNDRangeKernel() failed with %d\n", err);
	return NULL;
    }

    err = clEnqueueReadBufferRect(queues[pos], dev_res[pos], CL_FALSE, origin, origin, region_out, BASE_SIZE * sizeof(float), 0, 0, 0,  tmp_res[pos], 1, &ready, &cbevent);
    if (err != CL_SUCCESS) {
	printf("clEnqueueReadBufferRect() failed with %d\n", err);
	return NULL;
    }

    clSetEventCallback(cbevent, CL_COMPLETE, &exercise_block_finished_cb, info);

    clFlush(queues[pos]);

    return ready;
}

static size_t exercise_iteration(float *res, float *a, float *b, size_t size, int iterations) {
    size_t i;
    cl_event ready;

    blocks = size / BASE_SIZE;
    if (blocks * BASE_SIZE < size) ++blocks;

    last_block_size = size - (blocks - 1) * BASE_SIZE;
    current_block = 0;

    memset(res, 0, size * size * sizeof(float));

#ifdef DEBUG
    printf("Computing using %lux%lu blocks of size %lux%lu\n", blocks, blocks, BASE_SIZE, BASE_SIZE);
#endif

    runtime = 0;
    for (i = 0; i < PARALLEL; i++) {
	block_info[i].stopped = 0;
	block_info[i].pos = i;
	block_info[i].size = size;
	block_info[i].a = a;
	block_info[i].b = b;
	block_info[i].res = res;

	ready = exercise_schedule_block(&block_info[i]);
	if (ready) {
	    clSetUserEventStatus(ready, CL_COMPLETE);
	    clReleaseEvent(ready);
	} else {
#ifdef DEBUG
	    printf(" Skipping queue %lu\n", i);
#endif
	    block_info[i].stopped = 1;
	}
    }

    size_t finished; 
    do {
	finished = 0;
	for (i = 0; i < PARALLEL; i++) {
//	    clFinish(queues[i]);
	    if (block_info[i].stopped) finished++;
	}
    } while (finished < PARALLEL);

    if (current_block < blocks * blocks * blocks)
	return -1;

    return runtime;
}

size_t exercise(float *res, float *a, float *b, size_t size, int iterations) {
    int i;
    size_t ret;
    size_t runtime = 0;
    size_t missing = 0;
    
    for (i = 0; i < iterations; i++) {
	ret = exercise_iteration(res, a, b, size, iterations);
	if (ret == (size_t)-1) return ret;
	else if (ret == (size_t)-2) missing++;
	else runtime += ret;
    }

    if (missing) {
	if (missing == iterations) return 0;
	return 1.* iterations * runtime / (iterations - missing);
    }
    return runtime;
}

