#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <locale.h>

#define ITERATIONS 10
#define DEFAULT_SIZE 1000000

size_t exercise(float *res, float *a, float *b, size_t size, int iterations);
int exercise_init(const char *name, size_t size);
int exercise_allocate(float **a, float **b, size_t size);
void exercise_free();
extern int exercise_required_alignment;


size_t get_padded_size(size_t size, size_t block_size) {
    size_t blocks = (size / block_size);
    size_t inc = blocks * block_size - size;
    if (inc) return (blocks + 1) * block_size;
    return size;
}


int main(int argc, char *argv[]) {
    int err;
    long i;
    size_t size = DEFAULT_SIZE;
    size_t iterations = ITERATIONS;
    int standard_memory = 0;
    float *a = NULL, *b = NULL, res;

    size_t runtime;
    size_t us, throughput;
    struct timeval tv1,tv2;

    char *name, fname[255];
    FILE *f;

    name = strrchr(argv[0], '/');
    if (name) name += 1;
    else name = argv[0];

    if (argc > 1) {
        size = atoi(argv[1]);
    }
    
    if (argc > 2) {
	iterations = atoi(argv[2]);
	if (!iterations) iterations = 1;
    }

    if (exercise_required_alignment > 1) {
	size = get_padded_size(size, exercise_required_alignment);
    }

    throughput = 2 * size * sizeof(float);
    setlocale(LC_NUMERIC, "en_US");
    printf("Dot product of %'lu elements, Iterations %u, %lu Gthroughput required\n", size, iterations, (throughput/1000000000));

    err = exercise_init(name, size);
    if (err) return err;

    err = exercise_allocate(&a, &b, size);
    if (err) {
	if (res) return err;
	else printf("Allocation of fast memory failed, using standard slow mode...\n");
    }
    
    if (!a) {
	standard_memory = 1;
	
	a = (float *)calloc( size, sizeof( float ) );
        b = (float *)calloc( size, sizeof( float ) );
	if(a == NULL || b == NULL) {
    	    printf( "Can't allocate memory for arrays\n");
    	    return 0;
	}
    }

    srand(1);
    for (i = 0; i < size; i++) {
        a[i] = 1. * rand() / RAND_MAX;
        b[i] = 1. * rand() / RAND_MAX;
    }

    gettimeofday(&tv1, NULL);
    runtime = exercise(&res, a, b, size, iterations);
    gettimeofday(&tv2, NULL);

    if (runtime == (size_t)-1) {
	return 1;
    }
    
    printf("Result: %lf\n", res);

    exercise_free();
    
    if (standard_memory) {
	free(b);
	free(a);
    }

    us = (tv2.tv_sec - tv1.tv_sec)*1000000 + (tv2.tv_usec - tv1.tv_usec);
    us /= iterations;
    printf("Throughput: %.2lf GB/s (%lf s)\n", ((0.001 * throughput) / us), us / 1000000.);
    if (runtime) {
	runtime /= iterations;
        printf("Throughput (excluding transfer): %.2lf GB/s (%lf s)\n", ((1. * throughput) / runtime), runtime / 1000000000.);
    }

    return 0;
}

